package de.fraunhofer.iem.dataprovider.gitlab.service

import de.fraunhofer.iem.dataprovider.configuration.OpenCodeGitlabApiProperties
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.repository.dto.RepositoryCreateDto
import de.fraunhofer.iem.dataprovider.repository.dto.RepositoryDetailsDto
import kotlinx.coroutines.*
import org.gitlab4j.api.GitLabApi
import org.gitlab4j.api.models.Commit
import org.gitlab4j.api.models.Project
import org.springframework.stereotype.Service

/**
 * This class wraps all usages of org.gitlab4j and transforms their data models
 * into our internal data models.
 */
@Service
class OpenCodeGitlabApi(openCodeGitlabConfiguration: OpenCodeGitlabApiProperties) {
    private val logger = getLogger(javaClass)
    private val gitlabApi: GitLabApi =
        GitLabApi(openCodeGitlabConfiguration.host, openCodeGitlabConfiguration.accessToken)

    /**
     * Queries the gitlab project API at ${gitlabConfiguration.host} and returns a RepositoryCreateDto
     */
    fun getRepositoryInfo(projectId: Long): RepositoryCreateDto {
        logger.info("Get repository info for repository id $projectId started")

        val project = gitlabApi.projectApi.getProject(projectId)
        val projectUri = project.httpUrlToRepo
        val repoCreateDto = RepositoryCreateDto(project.path, projectUri, projectId)

        logger.info("Get repository info $repoCreateDto finished successfully.")
        return repoCreateDto
    }

    /**
     * Queries gitlab project, commit, and repository API to create a RepositoryDetailsDto
     */
    suspend fun getRepositoryDetails(projectId: Long): RepositoryDetailsDto {
        logger.info("Get repository details for repository id $projectId started")

        val project = gitlabApi.projectApi.getProject(projectId)
        // Note: We only take commits from the default branch
        val commits = gitlabApi.commitsApi.getCommits(project.id, project.defaultBranch, ".")

        val numberOfCommits = commits.count()
        val numberOfSignedCommits = getNumberOfSignedCommits(projectId, commits)
        val isDefaultBranchProtected = isDefaultBranchProtected(projectId, project)
        val repoDetailsDto = RepositoryDetailsDto(
            projectId,
            numberOfCommits,
            numberOfSignedCommits,
            isDefaultBranchProtected
        )

        logger.info("Get repository details $repoDetailsDto finished successfully.")
        return repoDetailsDto
    }

    private suspend fun getNumberOfSignedCommits(
        projectId: Long,
        commits: List<Commit>,
        ioDispatcher: CoroutineDispatcher = Dispatchers.IO
    ): Int =
        coroutineScope {
            val deferreds = commits.map {
                async(ioDispatcher) {
                    gitlabApi.commitsApi.getOptionalGpgSignature(projectId, it.id)
                }
            }

            deferreds.awaitAll().count { !it.isEmpty }
        }

    private fun isDefaultBranchProtected(projectId: Long, project: Project): Boolean {
        return try {
            val defaultBranchName = project.defaultBranch
            val branch = gitlabApi.repositoryApi.getBranch(projectId, defaultBranchName)
            branch.protected
        } catch (e: Exception) {
            // in theory, error probably happens if branch can't be found. In this case we default to false
            false
        }
    }
}
